include("../../pipeline.jl")


thisdir = Filesystem.abspath(@__DIR__)
projectname = Filesystem.dirname(thisdir)
base_prms_file = joinpath(thisdir, "base_parameters.toml")
base_prms = TOML.parsefile(base_prms_file)

### parameter variations
# every entry is a list of parameters and their values to vary
list_variations = []

# Standarad HRSC methods
basedict = Dict{String,Any}("type"=>"EntropyViscosity")
basedict["smoother"] = true
cmax = [ 1.0 ]
ce   = [ 0.1, 0.5, 1.0 ]
list = meshdict(; basedict=basedict, cmax, ce)
append!(list_variations, list)

basedict = Dict{String,Any}("type"=>"MDAViscosity")
basedict["smoother"] = true
cmax = [ 1.0 ]
list = meshdict(; basedict=basedict, cmax)
append!(list_variations, list)

basedict = Dict{String,Any}("type"=>"MDHViscosity")
basedict["smoother"] = true
cmax = [ 1.0 ]
ca = [ 0.1, 0.5, 1.0 ]
ck = [ 0.1, 0.5, 1.0 ]
list = meshdict(; basedict=basedict, cmax, ca, ck)
append!(list_variations, list)

basedict = Dict{String,Any}("type"=>"DBViscosity")
basedict["smoother"] = true
cmax = [ 1.0 ]
cb = [ 0.1, 0.5, 1.0 ]
list = meshdict(; basedict=basedict, cmax, cb)
append!(list_variations, list)

dict_list = [ merge(base_prms, Dict("HRSC"=>hrsc_prms)) for hrsc_prms in list_variations ]
for (i, d) in enumerate(dict_list)
  filename = joinpath(outputdir, "$(projectname)_$(i).toml")
  open(filename, "w") do file
    @info "Writing '$filename'..."
    TOML.print(file, d, sorted=true)
  end
end
