function make_callback(env, P, bdryconds)
  isperiodic = env.mesh.tree.periodic[1]
  cbfn_equation(u, t) = callback_equation(u, t, env, P, isperiodic, P.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P, isperiodic, P.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, isperiodic, P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  return callbackset
end


function make_callback(env, P::Project2d)
  # cbfn_equation(u, t) = callback_equation(u, t, env, P, true, P.equation)
  # cb_equation         = FunctionCallback(cbfn_equation,
  #                                        CallbackTiming(every_iteration=1,every_dt=0))
  # cbfn_tci(u, t)      = callback_tci(u, t, env, P, true, P.tci)
  # cb_tci              = FunctionCallback(cbfn_tci,
  #                                        CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, true, P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  # callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  callbackset = CallbackSet(cb_hrsc)
  return callbackset
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, env, P, isperiodic, eq::EulerEquation)
  @unpack cache = env
  @unpack equation = P
  wrap_dynamic_variables!(cache, state_u)
  @unpack Sm1, S, flx_S, flx_Sm1 = get_static_variables(cache)
  @unpack t, tm1 = get_global_variables(cache)
  tm1     .= t
  t       .= state_t
  Sm1     .= S
  flx_Sm1 .= flx_S
  dg1d.new_broadcast_volume!(pressure, equation, env.mesh)
  @unpack rho, q = get_dynamic_variables(cache)
  @unpack p, eps = get_static_variables(cache)
  dg1d.new_broadcast_volume!(entropy_variables, equation, env.mesh)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Nothing, mesh) end


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::AbstractReconstruction, mesh) = nothing


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Maybe{HRSC.AbstractHRSC}, mesh)
  TODO()
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P
  @unpack mu, max_v   = get_cell_variables(cache)
  @unpack v           = get_static_variables(cache)
  @unpack rho         = get_dynamic_variables(cache)
  dg1d.new_broadcast_volume!(maxspeed, equation, cache)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      rho, max_v,
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.TCIViscosity, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh1d)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache       = env
  @unpack mu          = get_cell_variables(mesh.cache)
  @unpack smoothed_mu = get_static_variables(mesh.cache)
  hrsc.smoother(smoothed_mu, mu, mesh, isperiodic)
end
function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh2d)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache       = env
  @unpack mu          = get_cell_variables(mesh.cache)
  @unpack smoothed_mu = get_static_variables(mesh.cache)
  hrsc.smoother(smoothed_mu, mu, mesh, #=isperiodic_x=#true, #=isperiodic_y=#true)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, ::Mesh1d)
  @unpack cache, mesh               = env
  @unpack rho                       = get_dynamic_variables(cache)
  @unpack S, Sm1, flx_S, flx_Sm1, v = get_static_variables(cache)
  @unpack mu, max_v                 = get_cell_variables(cache)
  @unpack t, tm1                    = get_global_variables(cache)
  t[1] <= 0.0 && tm1[1] <= 0 && return
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      rho, max_v, S, Sm1, flx_S, flx_Sm1, t[1], tm1[1],
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh2d)
  @unpack equation   = P
  @unpack S, Sm1, flx_S_x, flx_S_y,
          flxm1_S_x, flxm1_S_y, v,
          dxdX, dydY = get_static_variables(mesh.cache)
  @unpack mu         = get_cell_variables(mesh.cache)
  @unpack t, tm1     = get_global_variables(mesh.cache)

  # backup previous values
  @. Sm1   = S
  @. flxm1_S_x = flx_S_x
  @. flxm1_S_y = flx_S_y
  @. tm1   = t
  @. t     = state_t

  # compute new values
  dg1d.new_broadcast_volume!(entropy_variables, equation, mesh)
  dg1d.new_broadcast_volume!(maxspeed, equation, mesh)

  # TODO HRSC no longer provides an interface ...
  HRSC.compute_viscosity!(
      mu,
      S #= TODO This is unused =#,
      v, S, Sm1, flx_S_x, flx_S_y, flxm1_S_x, flxm1_S_y, t[1], tm1[1], dxdX, dydY,
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function callback_tci(state_u, state_t, env, P, isperiodic, tci::Nothing) end


function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.Threshold)
  error("Threshold TCI must be combined with an indicator that is supposed to be zero for " *
        "for smooth solutions, but not for shocks. TODO Implement an indicator variable.")
end


function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.EntropyProduction)
  @unpack cache, mesh                = env
  @unpack S, Sm1, flx_S, flx_Sm1, EP = get_static_variables(cache)
  @unpack flag                       = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs     = get_bdry_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         S, Sm1, flx_S, flx_Sm1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, env, P, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})

  @unpack cache, mesh = env
  @unpack rho         = get_dynamic_variables(cache)
  @unpack flag        = get_cell_variables(cache)
  TCI.compute_indicator!(flag, rho, tci)
end
