@doc """
    DBViscosity

Diffusion Based Artificial Viscosity model, cf. Yu, Hesthaven, 2019.
"""
Base.@kwdef struct DBViscosity <: AbstractArtificialViscosity
  mesh::Mesh
  cmax::Float64
  cb::Float64
end


function compute_viscosity!(
    mu,
    u, v,
    av::DBViscosity)

  @unpack mesh = av
  @unpack element, K, L, invjac = mesh
  @unpack D, N = element
  @unpack cmax, cb = av

  h = L / (K * N)

  mat_u = reshape(u, layout(mesh))
  maxabs_∇u = vec(maximum(abs.(D * mat_u .* invjac), dims=1))
  mu_b = cb * (h / N)^2 * maxabs_∇u

  maxabs_v = vec(maximum(abs.(v), dims=1))
  mu_max = @. cmax * h / N * maxabs_v

  @. mu = min(mu_b, mu_max)

  return true
end
