module SRHD


using dg1d


include("types.jl")
include("equation.jl")
include("rhs.jl")
include("initialdata.jl")
include("callbacks.jl")
include("boundaryconditions.jl")
include("setup.jl")


#######################################################################
#                 dg1d project interface definitions                  #
#######################################################################

dg1d.project_type(::Val{:SRHD}) = Project

# TODO Remove bdryconds
dg1d.rhs!(env::Environment, P::Project, bdryconds) = rhs!(env, P, P.hrsc, bdryconds...)
# dg1d.rhs!(env::Environment, P::Project2d, bdryconds) = rhs!(env, P, env.mesh)

dg1d.timestep!(env::Environment, P::Project) = timestep(env, P, P.hrsc)
dg1d.timestep!(env::Environment, P::Project2d) = timestep(env, P, nothing)

# TODO Can we combine dg1d.project_type definition and dg1d.@parameters
# dg1d.project_type(::Val{:SRHD}) = Project
dg1d.@parameters SRHD begin

  """
  Density of artificial atmosphere.
  """
  atm_density = 1e-10
  @check atm_density > 0.0

  """
  Multiplicative factor to determine the cutoff restmass density below which
  we impose artificial atmosphere.
  """
  atm_threshold_factor = 100.0
  @check atm_threshold_factor > 0.0

  """
  Initial data configuration. Available options:
  - `id_smooth`
  - `id_smooth_diagonal`
  - `id_periodic_relativistic_shocktube`
  - `id_periodic_relativistic_shocktube_2`
  - `id_periodic_shocktube_pressure_jump`
  - `id_periodic_shocktube_less_extreme`
  - `id_periodic_sod_shocktube`
  - `id_sod_shocktube`
  - `id_shocktube`
  """
  id = "id_smooth"
  # TODO I think the @check macro is not working properly...
  @check id in [ "id_smooth",
                 "id_smooth_diagonal",
                 "id_periodic_relativistic_shocktube",
                 "id_periodic_relativistic_shocktube_2",
                 "id_periodic_shocktube_pressure_jump",
                 "id_periodic_shocktube_less_extreme",
                 "id_periodic_sod_shocktube",
                 "id_sod_shocktube",
                 "id_shocktube_2d" ]

  "Flow speed for id_smooth test"
  id_smooth_flow_speed = 0.2
  @check id_smooth_flow_speed > 0.0

  """
  Boundary conditions. Available options:
  - `periodic`
  - `from_id`
  """
  bc = "periodic"
  @check bc in [ "periodic", "from_id" ]

end

end # module
