Base.@kwdef struct Atmosphere
  rho::Float64
  threshold::Float64
end


struct Equation{T_EoS<:EquationOfState.AbstractEquationOfState} <: AbstractEquation
  eos::T_EoS
  atmosphere::Atmosphere
  cons2prim_newtonsolver::dg1d.NewtonSolver
  cons2prim_bisection::dg1d.Bisection
end


struct Equation2d{T_EoS<:EquationOfState.AbstractEquationOfState} <: AbstractEquation
  eos::T_EoS
  atmosphere::Atmosphere
  cons2prim_newtonsolver::dg1d.NewtonSolver
  cons2prim_bisection::dg1d.Bisection
end


struct Project{T_RSolver  <:AbstractRiemannSolver,
               T_HRSC     <:Maybe{HRSC.AbstractHRSC},
               T_TCI      <:Maybe{TCI.AbstractTCI},
               T_LDG_RSolver<:Maybe{AbstractRiemannSolver},
               T_AV_RSolver<:Maybe{AbstractRiemannSolver}}

  equation::Equation
  rsolver::T_RSolver
  hrsc::T_HRSC
  tci::T_TCI
  ldg_rsolver::T_LDG_RSolver
  av_rsolver::T_AV_RSolver
end


struct Project2d{Eq<:Equation2d,
                 T_HRSC<:Maybe{HRSC.AbstractHRSC},
                 T_BC<:dg1d.AbstractBC}
  equation::Eq
  hrsc::T_HRSC
  bdrycond::T_BC
end
