__precompile__(false)
module dg1d

using Reexport
import Base: Filesystem, Threads
using Base64
using Jacobi
@reexport using HDF5
using LinearAlgebra
using LoopVectorization
import MacroTools
using OrderedCollections
@reexport import Parameters: @unpack, @with_kw
using PrettyTables
using Printf
using ProgressMeter
import Roots
using StaticArrays
using StructArrays
using TimerOutputs
@reexport using TOML
@reexport using ToggleableAsserts
using WriteVTK

#######################################################################
#                                utils                                #
#######################################################################

include("sandbox.jl")

export # parameter handling
       query, splitallext, trim_keys, match_keys, prepend_keys,
       # type utils
       concretetypes,
       # parameter files and file system operations
       mk_outputdir, dirname_path, flatten,
       # get_parameter_file, get_parameter_file_index,
       # select_parameter_files, gather_output_directories, gather_parameter_files,
       match_properties,
       # find_duplicated_parameter_files,
       # run_projects, analyze_projects,
       TODO, @TODO, @stringify_names,
       indent
include("utils.jl")


export LLF,
       absmax,
       periodize,
       estimate_convergence_rate,
       # deprecated
       periodic_index
include("numutils.jl")

export Output, read_variable, read_variable!, read_mesh
include("output.jl")

#######################################################################
#                            Numeric utils                            #
#######################################################################

export find_root
include("rootfinders.jl")
include("lagrange.jl")
include("glgl.jl")
include("lgl.jl")

#######################################################################
#                             Components                              #
#######################################################################

export SpectralElement
include("spectralelement.jl")

include("tensorbasis.jl")
include("box.jl")
include("tree.jl")
include("objectcache.jl")
# include("refvector.jl")

export Variables, n_variables, names
include("variablesdict.jl")
# choose between variables interface implementation
# const Variables = _VariablesArray
const Variables = _VariablesDict

export Cache, register_variables!,
       get_global_variables, get_dynamic_variables, get_cell_variables,
       get_bdry_variables, get_rhs_variables, get_static_variables,
       get_variable, wrap_dynamic_variables!, wrap_rhs_variables!,
       broadcast_volume!, variablenames, variablegroups,
       broadcast_volume_2!, broadcast_faces_2!, broadcast_bdry_2!
include("cache.jl")

export Mesh, Mesh1d, Mesh2d, MeshInterpolator, layout, n_points, n_cells, dimension,
       grid_average, cellwise_average, cellwise_inner_product, broken_inner_product,
       locate_point, find_cell_index, eachcell, cellindices
include("mesh.jl")

export AbstractEquation
include("equations.jl")

export XDMFOutput, xdmfwrite_mesh, xdmfwrite_scalar
include("xdmf.jl")

export VTKOutput, vtkwrite_scalar
include("vtk.jl")

export CallbackSet, CallbackTiming, FunctionCallback, SaveCallback, PlotCallback,
       TimeAlignedSaveCallback
include("callbacks.jl")

export AbstractRiemannSolver, ApproxRiemannSolver, ApproxRiemannSolver2d,
       lax_friedrich_flux, avg_lax_friedrich_flux,
       central_flux, mean_flux, broadcast_faces!
include("riemannsolver.jl")

export DirichletBC, OutflowBC, InflowBC, PeriodicBC, BoundaryConditions, isperiodic,
       broadcast_boundaryconditions!
include("bdryconditions.jl")

export @with_signature, @accepts, @returns,
       has_signature, accepts, returns, signature, state_indices
include("with_signature.jl")

export compute_rhs_weak_form!
include("dg_rhs.jl")

# TODO Move this to top; also requires gather all type defintions
# like SpectralElement, AbstractMesh, Tree etc. into types.jl
export Maybe, Environment
include("types.jl")

include("evolve.jl")

#######################################################################
#                              Projects                               #
#######################################################################

include("projects.jl")

export parameters, @parameters
include("parameters.jl")

export main, testparfiles
include("main.jl")

include("TCI/TCI.jl")
@reexport using .TCI
include("HRSC/HRSC.jl")
@reexport using .HRSC
include("EquationOfState/EquationOfState.jl")
@reexport using .EquationOfState
include("ScalarEq/ScalarEq.jl")
include("HeatEq/HeatEq.jl")
include("EulerEq/EulerEq.jl")
include("SRHD/SRHD.jl")
# include("GRHD/GRHD.jl")


end # module
